#define BITMAP_ID 0x4D42

/*******************************************************************/
// LoadBMPFile(char*)
/*******************************************************************/

bitmap* LoadBMPFile(char *file_name)
{
	FILE *bmp_file;
	BITMAPFILEHEADER bitmapfileheader;
	BITMAPINFOHEADER bitmapinfoheader;
	bitmap *new_bitmap;

	bmp_file = fopen(file_name, "rb");

	if (bmp_file != NULL) {
		fread(&bitmapfileheader, sizeof(BITMAPFILEHEADER), 1, bmp_file);

		if (bitmapfileheader.bfType == BITMAP_ID) {
			fread(&bitmapinfoheader, sizeof(BITMAPINFOHEADER), 1, bmp_file);

			fseek(bmp_file, bitmapfileheader.bfOffBits, SEEK_SET);

			new_bitmap = new bitmap(bitmapinfoheader.biWidth, bitmapinfoheader.biHeight, bitmapinfoheader.biBitCount / 8);

			if (new_bitmap != NULL) {
				fread(new_bitmap->bitmap_data, 1, bitmapinfoheader.biSizeImage, bmp_file);

				for (unsigned int index = 0; index < bitmapinfoheader.biSizeImage; index += 3) {
					unsigned char temp;

					temp = new_bitmap->bitmap_data[index];
					new_bitmap->bitmap_data[index] = new_bitmap->bitmap_data[index+2];
					new_bitmap->bitmap_data[index+2] = temp; }

				new_bitmap->y = bitmapinfoheader.biHeight;
				new_bitmap->x = bitmapinfoheader.biWidth;
				new_bitmap->planes = bitmapinfoheader.biBitCount / 8;
			} else cout << "ERROR: Could not allocate memory for bitmap!" << endl;
		}else cout << "ERROR: Not a valid BMP file!" << endl;

		fclose(bmp_file);
	} else cout << "ERROR: File '" << file_name << "' could not be opened!" << endl;

	return new_bitmap;
}

/*******************************************************************/
// WriteBMPFile(char*, bitmap*)
/*******************************************************************/

unsigned char WriteBMPFile(char *file_name, bitmap *old_bitmap)
{
	FILE *bmp_file;
	BITMAPFILEHEADER bitmapfileheader;
	BITMAPINFOHEADER bitmapinfoheader;
	unsigned char *bitmap_data, success = NULL;

	if (old_bitmap != NULL) {
		if (old_bitmap->bitmap_data != NULL) {

			bmp_file = fopen(file_name, "wb");

			if (bmp_file != NULL) {
				int bitmap_size = old_bitmap->x * old_bitmap->y * old_bitmap->planes;

				bitmapfileheader.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + bitmap_size;
				bitmapfileheader.bfType = BITMAP_ID;
				bitmapfileheader.bfReserved1 = 0;
				bitmapfileheader.bfReserved2 = 0;
				bitmapfileheader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);

				bitmapinfoheader.biSize = sizeof(BITMAPINFOHEADER);
				bitmapinfoheader.biPlanes = 1;
				bitmapinfoheader.biBitCount = old_bitmap->planes * 8;
				bitmapinfoheader.biCompression = BI_RGB;
				bitmapinfoheader.biSizeImage = bitmap_size;
				bitmapinfoheader.biXPelsPerMeter = 0;
				bitmapinfoheader.biYPelsPerMeter = 0;
				bitmapinfoheader.biClrUsed = 0;
				bitmapinfoheader.biClrImportant = 0;
				bitmapinfoheader.biWidth = old_bitmap->x;
				bitmapinfoheader.biHeight = old_bitmap->y;

				bitmap_data = new unsigned char [bitmap_size];

				if (bitmap_data != NULL) {

					if (old_bitmap->planes == 3) {
						for (int index = 0; index < bitmap_size; index += 3) {
							bitmap_data[index] = old_bitmap->bitmap_data[index + 2];
							bitmap_data[index + 1] = old_bitmap->bitmap_data[index + 1];
							bitmap_data[index + 2] = old_bitmap->bitmap_data[index]; } }
					else {
						for (int index = 0; index < bitmap_size; index++) {
							bitmap_data[index] = old_bitmap->bitmap_data[index]; } }
					
					fwrite(&bitmapfileheader, 1, sizeof(BITMAPFILEHEADER), bmp_file);
					fwrite(&bitmapinfoheader, 1, sizeof(BITMAPINFOHEADER), bmp_file);
					fwrite(bitmap_data, 1, bitmap_size, bmp_file);

					fclose(bmp_file);

					success = 1;
				} else cout << "ERROR:Could not allocate enough memory for temporary bitmap!" << endl;
			} else cout << "ERROR: Could not open BMP file '" << file_name << "' for writing!" << endl;
		} else cout << "ERROR: Bitmap is empty!" << endl;
	} else cout << "ERROR: Pointer to bitmap is NULL!" << endl;
	return success;
}

/*******************************************************************/
// PrintBMPFileData(char*)		:Displays information about a BMP file!
/*******************************************************************/

void PrintBMPFileData(char *file_name)
{
	BITMAPFILEHEADER bfh;
	BITMAPINFOHEADER bih;
	FILE *bmp_file;

	bmp_file = fopen(file_name, "rb");

	if (bmp_file != NULL)
	{
		fread(&bfh, 1, sizeof(BITMAPFILEHEADER), bmp_file);
		fread(&bih, 1, sizeof(BITMAPINFOHEADER), bmp_file);

		cout << "Info for bitmap file: " << file_name << endl << endl;
		cout << "bfSize:          " << bfh.bfSize << endl;
		cout << "bfType:          " << bfh.bfType << endl;
		cout << "bfReserved1:     " << bfh.bfReserved1 << endl;
		cout << "bfReserved2:     " << bfh.bfReserved2 << endl;
		cout << "bfOffBits:       " << bfh.bfOffBits << endl;
		cout << endl;
		cout << "biSize:          " << bih.biSize << endl;
		cout << "biPlanes:        " << bih.biPlanes << endl;
		cout << "biBitCount:      " << bih.biBitCount << endl;
		cout << "biCompression:   " << bih.biCompression << endl;
		cout << "biSizeImage:     " << bih.biSizeImage << endl;
		cout << "biXPelsPerMeter: " << bih.biXPelsPerMeter << endl;
		cout << "biYPelsPerMeter: " << bih.biYPelsPerMeter << endl;
		cout << "biClrUsed:       " << bih.biClrUsed << endl;
		cout << "biClrImportant:  " << bih.biClrImportant << endl;
		cout << "biWidth:         " << bih.biWidth << endl;
		cout << "biHeight:        " << bih.biHeight << endl;
		cout << endl;

		fclose(bmp_file);
	} else cout << "ERROR: Could not load BMP file '" << file_name << "'!" << endl;
}

/*******************************************************************/
//
/*******************************************************************/
