// light.h: interface for the light class.
//
//////////////////////////////////////////////////////////////////////

#ifndef _LIGHT_H_
#define _LIGHT_H_

#include "vector.h"
//#include "object.h"

class object;

class light  
{
public:
	light();
	virtual ~light();

	virtual vector getLightDirection(vector);
	virtual colour getLightIntensity(vector);

	virtual void setPosition(vector);
	virtual void setDirection(vector);
	virtual void setAttenuation(double);
	void setIntensity(colour);

	virtual void generateShadowMap(object**, int);
	virtual void generateShadowVolume(object**, int);
	virtual void drawLight(void);


	light *nextptr;

protected:
	colour intensity;
};

class pointlight : public light
{
public:
	pointlight();
	virtual ~pointlight();

	vector getLightDirection(vector);
	colour getLightIntensity(vector);

	virtual void setPosition(vector);
	virtual void setAttenuation(double);

	void drawLight(void);
	void generateShadowVolume(object**, int);

	vector position;

	double linear_attenuation;
};

class directedlight : public light
{
public:
	directedlight();
	virtual ~directedlight();

	vector getLightDirection(vector);
	colour getLightIntensity(vector);

	virtual void setDirection(vector);
	virtual void generateShadowVolume(object**, int);

	void drawLight(void);

private:
	vector direction;
};

class spotlight : public light
{
public:
	spotlight();
	virtual ~spotlight();

	vector getLightDirection(vector);
	colour getLightIntensity(vector);

	virtual void setPosition(vector);
	virtual void setDirection(vector);

	void drawLight(void);

	void generateShadowMap(object**, int);
	virtual void generateShadowVolume(object**, int);

	matrix modelview_matrix;
	matrix projection_matrix;
	matrix shadow_matrix;

	double umbra;
	double penumbra;

	vector position;
	vector direction;

	double shadow_map_x;
	double shadow_map_y;
	float *shadow_map;

	double max_distance;
	double min_distance;

	double smallest;
	double largest;
};


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

light::light()
{
	intensity.r = 255.0;
	intensity.g = 255.0;
	intensity.b = 255.0;

	nextptr = NULL;
}

pointlight::pointlight() : light()
{
	position.x = 10.0;
	position.y = 10.0;
	position.z = 10.0;

	linear_attenuation = -1;
}

directedlight::directedlight() : light()
{
	direction.x = -1.0;
	direction.y = -1.0;
	direction.z = -1.0;
}

spotlight::spotlight() : light()
{
	position.x = 0.0;
	position.y = 0.0;
	position.z = 0.0;

	direction.x = 1.0;
	direction.y = -1.0;
	direction.z = -1.0;

	umbra = 0.7;
	penumbra = 0.6;

	max_distance = 40.0;
	min_distance = 10.0;

	shadow_map_x = 256;
	shadow_map_y = 256;

	shadow_map = NULL;
}


light::~light()
{

}

directedlight::~directedlight()
{

}

pointlight::~pointlight()
{

}

spotlight::~spotlight()
{

}
#endif
