// object.cpp: implementation of the object class.
//
//////////////////////////////////////////////////////////////////////

#include <string.h>
#include "object.h"

#define FALSE 0
#define TRUE 1

// Flags to determine whether or not to calculate the lighting textures

unsigned char CALC_AMBIENT_MAP = TRUE;
unsigned char CALC_DIFFUSE_MAP = TRUE;
unsigned char CALC_SPECULAR_MAP = FALSE;

unsigned char CALC_AMBIENT_MAP_REALTIME = FALSE;
unsigned char CALC_DIFFUSE_MAP_REALTIME = FALSE;
unsigned char CALC_SPECULAR_MAP_REALTIME = FALSE;

unsigned char EFFECT_ANTIALIASING = FALSE;
unsigned char EFFECT_DEPTH_OF_FIELD = TRUE;

colour _GLOBAL_AMBIENT(100, 100, 100, 255);
vector _CAMERA;
light *_LIGHTS;

int percent(int a, int b)
{
	float c = (float)(100.0 / b);
	return (int)(c * a);
}

/******************************************************************************/
// Generates the polygon map. Each pixel in the texture is mapped to a point in
// UV space, then that point is tested as to whether or not it is inside a UV polygon
// If it is, then a pointer to that polygon is stored in the map, otherwise NULL is
// stored!
/******************************************************************************/

void object::calcAlphaMap(void)
{
#ifdef _DEBUG_OBJ
	cout << "calcAlphaMap(void)" << endl;
#endif

	double x_step, y_step;
	int x, y, count, max, last(-1);
	vector pos;
	colour w(255, 255, 255, 255), b(0, 0, 0, 0);

	x = final_colour->x;
	y = final_colour->y;

	x_step = 1.0 / x;
	y_step = 1.0 / y;

	count = 0;
	max = x * y * polygon_count;

	for (int j = 0; j < y; j++) {
		pos.y = y_step * j + (y_step / 2.0);
		for (int i = 0; i < x; i++) {
			pos.x = x_step * i + (x_step / 2.0);
			int istrue = 0;
			vector pos2 = pos;
			pos2.x -= x_step / 2.0;
			pos2.y -= y_step / 2.0;

			if (last != -1)
			{
				if (isPointInsidePolygon(pos, shading_polygons[last].uv1, shading_polygons[last].uv2, shading_polygons[last].uv3) == 1)
				{
					istrue = 1;
				} else {
					last = -1;
				}
			}

			if (istrue == 0)
			{
				if (last != -1)
				{
					if (isPointInsidePolygon(pos2, shading_polygons[last].uv1, shading_polygons[last].uv2, shading_polygons[last].uv3) == 1)
					{
						istrue = 1;
					} else {
						last = -1;
					}
				}
			}
			
			for (int k = 0; k < polygon_count && istrue != 1; k++) {
				if (isPointInsidePolygon(pos, shading_polygons[k].uv1, shading_polygons[k].uv2, shading_polygons[k].uv3) == 1)
				{
					istrue = 1;
					last = k;
				}
				else if (isPointInsidePolygon(pos2, shading_polygons[k].uv1, shading_polygons[k].uv2, shading_polygons[k].uv3) == 1)
				{
					istrue = 1;
					last = k;
				}
			}					

			if (istrue == 1) alpha_map[i][j] = &(shading_polygons[last]);
		}
		count += polygon_count * y;
		cout << "Generating Alpha Map: " << percent(count, max) << "\%\r";
	}

// To reduce the alising effect, empty points are mapped to their neighbouring points polygons

	for (int j = 0; j < y; j++) {
		for (int i = 0; i < x; i++) {
			if (alpha_map[i][j] == NULL)
			{
				if (i < x - 1)
				{
					if (alpha_map[i + 1][j] != NULL) alpha_map[i][j] = alpha_map[i + 1][j];
					else {
						if (i > 0)
						{
							if (alpha_map[i - 1][j] != NULL) alpha_map[i][j] = alpha_map[i - 1][j];
						}
					}
				}
			}

			if (alpha_map[i][j] == NULL)
			{
				if (j < y - 1)
				{
					if (alpha_map[i][j + 1] != NULL) alpha_map[i][j] = alpha_map[i][j + 1];
					else {
						if (j > 0)
						{
							if (alpha_map[i][j - 1] != NULL) alpha_map[i][j] = alpha_map[i][j - 1];
						}
					}
				}
			}
		}
	}
		
	cout << endl;
}

/******************************************************************************/
// Calculates the positions map. It finds the point in world space that corresponds
// to the point on the texture
/******************************************************************************/

void object::calcPositionsMap(void)
{
#ifdef _DEBUG_OBJ
	cout << "calcPositionsMap(void)" << endl;
#endif

	double x_step, y_step;
	int x, y, count, max;
	vector pos, v;
	colour w(255, 255, 255, 255);

	x = final_colour->x;
	y = final_colour->y;

	x_step = 1.0 / x;
	y_step = 1.0 / y;

	count = 0;
	max = x * y * polygon_count;

	for (int j = 0; j < y; j++) {
		pos.y = y_step * j;
		for (int i = 0; i < x; i++) {
			polygon *cur_polygon;
			pos.x = x_step * i;

			cur_polygon = alpha_map[i][j];

			if (cur_polygon != NULL) v = cur_polygon->uvSpaceToWorldSpace(pos);
			else v(0.0, 0.0, 0.0);

			position_map[j][i] = v;
		}
		count += polygon_count * y;
		cout << "Generating Positions Map: " << percent(count, max) << "\%\r";
	}
	cout << endl;
}

/******************************************************************************/
// Calculates the normals map. It finds the normal on the object that corresponds
// to the point on the texture
/******************************************************************************/

void object::calcNormalsMap(void)
{
#ifdef _DEBUG_OBJ
	cout << "calcNormalsMap(void)" << endl;
#endif

	double x_step, y_step;
	int x, y, count, max;
	vector pos, v;
	colour w(255, 255, 255, 255);

	x = final_colour->x;
	y = final_colour->y;

	x_step = 1.0 / x;
	y_step = 1.0 / y;

	count = 0;
	max = x * y * polygon_count;

	for (int j = 0; j < y; j++) {
		pos.y = y_step * j;
		for (int i = 0; i < x; i++) {
			polygon *cur_polygon;

			pos.x = x_step * i;

			cur_polygon = alpha_map[i][j];

			if (cur_polygon != NULL)
			{
				v = cur_polygon->getNormalAtUV(pos);
				v.normalise();

				if (bump_map != NULL)
				{
					vector tangent1, tangent2, up(0.0, 1.0, 0.0);
					double bu, bv;
					colour a, b, c;

					if (v != up)
					{
						tangent1 = v % up;
						up = tangent1 % v;
					}
					else {
						up.x = 1.0;
						up.y = 0.0;
						up.z = 0.0;

						tangent1 = up % v;
						up = v % tangent1;
					}

					a = bump_map->getColourf((double)pos.x, (double)pos.y);
					b = bump_map->getColourf((double)pos.x - x_step, (double)pos.y);
					c = bump_map->getColourf((double)pos.x, (double)pos.y - y_step);

					bu = (1.0 / 255.0) * (a.r - b.r);
					bv = (1.0 / 255.0) * (a.r - c.r);

				//	cout << bu << ", " << bv << endl;

					v = v + (tangent1 * bv) + (up * bu);

					v.normalise();
				}
			}
			else v(0.0, 0.0, 0.0);

			normals_map[j][i] = v;
		}
		count += polygon_count * y;
		cout << "Generating Normals Map: " << percent(count, max) << "\%\r";
	}
	cout << endl;
}

#define DIFFUSE		0
#define AMBIENT		1
#define SPECULAR	2
#define FINAL		3
#define ALPHA		4
#define NORMALS		5
#define POSITIONS	6

/******************************************************************************/
// Sets the material properties of the object to a flat colour
/******************************************************************************/

void object::setFlatColour(int attrib, colour col)
{
	colour *flat;

	flat = new colour;

	flat->r = col.r;
	flat->g = col.g;
	flat->b = col.b;
	flat->a = col.a;

	switch (attrib) {
	case DIFFUSE:
		if (diffuse.col != NULL) {
			delete [] diffuse.col;
			diffuse.col = NULL; }
		if (diffuse.bit != NULL) {
			delete [] diffuse.bit;
			diffuse.bit = NULL; }
		diffuse.col = flat;
		break;
	case AMBIENT:
		if (ambient.col != NULL) {
			delete [] ambient.col;
			ambient.col = NULL; }
		if (ambient.bit != NULL) {
			delete [] ambient.bit;
			ambient.bit = NULL; }
		ambient.col = flat;
		break;
	case SPECULAR:
		if (specular.col != NULL) {
			delete [] specular.col;
			specular.col = NULL; }
		if (specular.bit != NULL) {
			delete [] specular.bit;
			specular.bit = NULL; }
		specular.col = flat;
		break;
	default:
		break;
	}
}

/******************************************************************************/
// Sets the material properties of an object to a texture
/******************************************************************************/

void object::setColourTexture(int attrib, bitmap bit)
{
	bitmap *tex;

#ifdef _DEBUG_OBJ
	cout << "setColourTexture(" << attrib << ");" << endl;
#endif

	tex = new bitmap;

	tex->x = bit.x;
	tex->y = bit.y;
	tex->planes = bit.planes;
	tex->bitmap_data = bit.bitmap_data;

	cout << "Setting ";

	switch (attrib) {
	case DIFFUSE:
		cout << "DIFFUSE";
		if (diffuse.col != NULL) {
			delete [] diffuse.col;
			diffuse.col = NULL; }
		if (diffuse.bit != NULL) {
			delete [] diffuse.bit;
			diffuse.bit = NULL; }
		diffuse.bit = tex;
		break;
	case AMBIENT:
		cout << "AMBIENT";
		if (ambient.col != NULL) {
			delete [] ambient.col;
			ambient.col = NULL; }
		if (ambient.bit != NULL) {
			delete [] ambient.bit;
			ambient.bit = NULL; }
		ambient.bit = tex;
		break;
	case SPECULAR:
		cout << "SPECULAR";
		if (specular.col != NULL) {
			delete [] specular.col;
			specular.col = NULL; }
		if (specular.bit != NULL) {
			delete [] specular.bit;
			specular.bit = NULL; }
		specular.bit = tex;
		break;
	default:
		break;
	}
	cout << " to bitmap texture" << endl;
}

/******************************************************************************/
// Sets the objects bump map
/******************************************************************************/

void object::setBumpMap(bitmap *new_bump_map)
{
	bump_map = new_bump_map;
}

/******************************************************************************/
// Sets the objects colour map
/******************************************************************************/

void object::setColourMap(bitmap *new_colour_map)
{
	colour_map = new_colour_map;
}

/******************************************************************************/
// Initialises the shading bitmaps to all 0's
/******************************************************************************/

void object::initialiseBitmaps(int attrib)
{
#ifdef _DEBUG_OBJ
	cout << "initialiseBitmaps(" << attrib << ");" << endl;
#endif

	unsigned char *data;
	int size;

	switch (attrib) {
	case DIFFUSE:
		cout << "Initialising bitmap : DIFFUSE" << endl;
		size = diffuse_colour->x * diffuse_colour->y * diffuse_colour->planes;
		data = diffuse_colour->bitmap_data;
		for(int i = 0; i < size; i++) data[i] = 0;
		break;
	case AMBIENT:
		cout << "Initialising bitmap : AMBIENT" << endl;
		size = ambient_colour->x * ambient_colour->y * ambient_colour->planes;
		data = ambient_colour->bitmap_data;
		for(int i = 0; i < size; i++) data[i] = 0;
		break;
	case SPECULAR:
		cout << "Initialising bitmap : SPECULAR" << endl;
		size = specular_colour->x * specular_colour->y * specular_colour->planes;
		data = specular_colour->bitmap_data;
		for(int i = 0; i < size; i++) data[i] = 0;
		break;
	case FINAL:
		cout << "Initialising bitmap : FINAL COLOUR" << endl;
		size = final_colour->x * final_colour->y * final_colour->planes;
		data = final_colour->bitmap_data;
		for(int i = 0; i < size; i++) data[i] = 0;
		break;
	default:
		cout << "ERROR: No bitmap specified to initialise!" << endl;
		break;
	}
}

/******************************************************************************/
// Calculates the Ambient properties of the object and stores in a texture
/******************************************************************************/

void object::calcAmbientMap(colour amb)
{
#ifdef _DEBUG_OBJ
	cout << "calcAmbientMap(colour);" << endl;
#endif

	int x, y, p, i, j, index;
	unsigned char *data;
	colour c;
	float ra, ga, ba, n;
	polygon *cur_polygon;

	n = 1.0 / 255.0;

	ra = amb.r * n;
	ga = amb.g * n;
	ba = amb.b * n;

	x = ambient_colour->x;
	y = ambient_colour->y;
	p = ambient_colour->planes;
	data = ambient_colour->bitmap_data;

	for (j = 0; j < y; j++) {
		for (i = 0; i < x; i++) {
			cur_polygon = alpha_map[i][j];
			if (cur_polygon != NULL) {
				if (cur_polygon->visible == 1)
				{
					index = (j * y * p) + (i * p);
					c = ambient.getColour(i, j);
					data[index] = c.r * ra;
					data[index+1] = c.g * ga;
					data[index+2] = c.b * ba;
	} } } }
}

/******************************************************************************/
// Calculates the diffuse properties of the object and stores in a texture
/******************************************************************************/

void object::calcDiffuseMap(light *lights, vector eye)
{
#ifdef _DEBUG_OBJ
	cout << "calcDiffuseMap(colour);" << endl;
#endif
	light *cur_light;
	int x, y, p, i, j, index;
	unsigned char *data;
	colour c, in;
	float ra, ga, ba, rc, gc, bc, n, d, lr, lg, lb;
	polygon *cur_polygon;
	vector pos, ray, norm;

	n = 1.0 / 255.0;

	x = diffuse_colour->x;
	y = diffuse_colour->y;
	p = diffuse_colour->planes;

	data = diffuse_colour->bitmap_data;

	for (i = 0; i < x; i++) {
		for (j = 0; j < y; j++) {
			cur_polygon = alpha_map[i][j];
			if (cur_polygon != NULL) {
				if (cur_polygon->visible == 1)
				{
					pos = position_map[j][i];
					norm = normals_map[j][i];
					
					c = diffuse.getColour(i, j);
					
					lr = 0.0;
					lg = 0.0;
					lb = 0.0;
					
					cur_light = lights;
					
					while (cur_light != NULL) {
						ray = cur_light->getLightDirection(pos);
						ray.normalise();
						in = cur_light->getLightIntensity(pos);
						d = dot(ray, norm);
						if (d > 0.0) {
							lr = lr + ((in.r * n) * d);
							lg = lg + ((in.g * n) * d);
							lb = lb + ((in.b * n) * d); }
			
						cur_light = cur_light->nextptr;
					}
					index = (j * y * p) + (i * p);
					
					if (lr > 1.0) lr = 1.0;
					if (lg > 1.0) lg = 1.0;
					if (lb > 1.0) lb = 1.0;
					
					data[index] = c.r * lr;
					data[index + 1] = c.g * lg;
					data[index + 2] = c.b * lb;
				}
			}
		}
	}
}

/******************************************************************************/
// Calculates ALL material properties of an object at the same time (faster than doing all
// seperately
/******************************************************************************/

void object::calcAllMaps(light *lights, vector eye, colour a)
{
#ifdef _DEBUG_OBJ
	cout << "calcAllMaps(colour);" << endl;
#endif

	light *cur_light;
	int x, y, p, i, j, index;
	unsigned char *data;
	colour c, in, amb, dif, spec;
	float dlr, dlg, dlb, slr, slg, slb, n, d, lr, lg, lb;
	polygon *cur_polygon;
	vector pos, ray, norm, eye_ray, h;

	n = 1.0 / 255.0;

	x = final_colour->x;
	y = final_colour->y;
	p = final_colour->planes;

	data = final_colour->bitmap_data;

	for (i = 0; i < x; i++) {
		for (j = 0; j < y; j++) {
			cur_polygon = alpha_map[i][j];
			if (cur_polygon != NULL) {
				if (cur_polygon->visible == 1)
				{
					pos = position_map[j][i];
					norm = normals_map[j][i];
					
					eye_ray = eye - pos;
					
					amb = ambient.getColour(i, j);
					dif = diffuse.getColour(i, j);
					spec = specular.getColour(i, j);
					
					slr = 0.0;
					slg = 0.0;
					slb = 0.0;
					
					dlr = 0.0;
					dlg = 0.0;
					dlb = 0.0;
					
					cur_light = lights;
					
					while (cur_light != NULL) {
						ray = cur_light->getLightDirection(pos);
						ray.normalise();
						in = cur_light->getLightIntensity(pos);
						
						d = dot(ray, norm);
						
						if (d > 0.0)
						{
							dlr += (in.r * n) * d;
							dlg += (in.g * n) * d;
							dlb += (in.b * n) * d;
							
							if (dot(norm, ray) > 0.0)
							{
								eye_ray.normalise();
								
								h = ray + eye_ray;
								h.normalise();
								
								d = pow(dot(norm, h), specular_coefficient);
								
								slr += (in.r * n) * d;
								slg += (in.g * n) * d;
								slb += (in.b * n) * d;
							}
						}
						
						cur_light = cur_light->nextptr;
					}
					index = (j * y * p) + (i * p);
					
					lr = (dlr * dif.r) + (slr * spec.r) + (amb.r * (a.r * n));
					lg = (dlg * dif.g) + (slg * spec.g) + (amb.g * (a.g * n));
					lb = (dlb * dif.b) + (slb * spec.b) + (amb.b * (a.b * n));
					
					if (lr > 255.0) lr = 255.0;
					if (lg > 255.0) lg = 255.0;
					if (lb > 255.0) lb = 255.0;
					
					data[index] = lr;
					data[index + 1] = lg;
					data[index + 2] = lb;
				}
			}
		}
	}
}

/******************************************************************************/
// Calculates the specular properties of an object and stores in a texture
/******************************************************************************/

void object::calcSpecularMap(light *lights, vector eye)
{
#ifdef _DEBUG_OBJ
	cout << "calcSpecularMap(colour);" << endl;
#endif

	light *cur_light;
	int x, y, p, i, j, index;
	unsigned char *data;
	colour c, in;
	float ra, ga, ba, rc, gc, bc, n, d, lr, lg, lb;
	polygon *cur_polygon;
	vector pos, ray, norm, eye_ray, h;

	n = 1.0 / 255.0;

	x = specular_colour->x;
	y = specular_colour->y;
	p = specular_colour->planes;

	data = specular_colour->bitmap_data;

	for (i = 0; i < x; i++) {
		for (j = 0; j < y; j++) {
			cur_polygon = alpha_map[i][j];
			if (cur_polygon != NULL) {
				if (cur_polygon->visible == 1)
				{
					pos = position_map[j][i];
					norm = normals_map[j][i];
					eye_ray = eye - pos;
					
					c = specular.getColour(i, j);
					
					lr = 0.0;
					lg = 0.0;
					lb = 0.0;
					
					cur_light = lights;
					
					while (cur_light != NULL) {
						ray = cur_light->getLightDirection(pos);
						
						if (dot(norm, ray) > 0.0)
						{
							eye_ray.normalise();
							ray.normalise();
							
							in = cur_light->getLightIntensity(pos);
							h = ray + eye_ray;
							h.normalise();

							d = pow(dot(norm, h), specular_coefficient);
							
							lr = lr + ((in.r * n) * d);
							lg = lg + ((in.g * n) * d);
							lb = lb + ((in.b * n) * d);
						}
						
						cur_light = cur_light->nextptr;
					}
					index = (j * y * p) + (i * p);
					
					if (lr > 1.0) lr = 1.0;
					if (lg > 1.0) lg = 1.0;
					if (lb > 1.0) lb = 1.0;
					
					data[index] = c.r * lr;
					data[index + 1] = c.g * lg;
					data[index + 2] = c.b * lb;
				}
			}
		}
	}
}

/******************************************************************************/
// Merges the Ambient, Diffuse and Specular properties together into one final texture
/// *** THIS IS THE TEXTURE THAT WILL BE USED
/******************************************************************************/

void object::combineColourComponents(void)
{
#ifdef _DEBUG_OBJ
	cout << "combineColourComponents(void);" << endl;
#endif

	int i, max, c;

	max = final_colour->x * final_colour->y * final_colour->planes;

	for (i = 0; i < max; i++) {
		c = ambient_colour->bitmap_data[i]
			+ diffuse_colour->bitmap_data[i]
			+ specular_colour->bitmap_data[i];
		final_colour->bitmap_data[i] = (c > 255 ? 255 : c < 0 ? 0 : c); }
}

/******************************************************************************/
// Tests whether each polygon of an object is visible in the camera
// Uses the polygons plane to determine which side of the polygon the camera is
// on. If on the front, the polygon is visible, else it's not
/******************************************************************************/

void object::visibilityTest(vector eye_pos)
{

#ifdef _DEBUG_OBJ
	cout << "visibilityTest(" << eye_dir << ");" << endl;
#endif

	int k;
	vector v;

	for (k = 0; k < polygon_count; k++)
	{
		v = eye_pos - polygons[k].vertex1;
		if (dot(polygons[k].plane_normal, v) > 0.0)
		{
			polygons[k].visible = 1;
		} else polygons[k].visible = 0;
	}
}

/******************************************************************************/
// In order to successfully determine whether or not a point is in the polygon,
// the UVs of the polygon must be counter-clockwise. This test determines whether
// they are or not. If not, then it reverses them
/******************************************************************************/

void object::UVCounterClockwiseTest(void)
{
#ifdef _DEBUG_OBJ
	cout << "UVCounterClockwiseTest(void);" << endl;
#endif

	cout << endl << "PERFORMING UV COUNTER CLOCKWISE TEST!" << endl;

	int i;
	vector v1, v2, n, temp;

	for (i = 0; i < polygon_count; i++)
	{
		v1 = polygons[i].uv2 - polygons[i].uv1;
		v2 = polygons[i].uv3 - polygons[i].uv1;
		n = v1 % v2;

		if (n.z < 0.0) 
		{
			temp = polygons[i].uv2;
			polygons[i].uv2 = polygons[i].uv3;
			polygons[i].uv3 = temp;
			i--;
		}
	}
}

/******************************************************************************/
// Some UV's stretch outside of the 0 - 1 bounds of the texture and are either tiled
// or clamped usually, this may work with colour textures, but it would look wrong with
// shading, so all UV's are scaled to fit in this 0 - 1 limits
/******************************************************************************/

void object::scaleUVs(void)
{
	double minx, miny, maxx, maxy;

	cout << endl << "SCALING OBJECTS UVS TO WITHIN 0 - 1 LIMITS!" << endl;

	if (uv_count > 0) {
		minx = uvs[0].x;
		miny = uvs[0].y;
		maxx = uvs[0].x;
		maxy = uvs[0].y;

		for (int i = 0; i < uv_count; i++)
		{
			if (uvs[i].x < minx) minx = uvs[i].x;
			if (uvs[i].y < miny) miny = uvs[i].y;

			if (uvs[i].x > maxx) maxx = uvs[i].x;
			if (uvs[i].y > maxy) maxy = uvs[i].y;
		}

		double scalex, scaley;
		
		scalex = 1.0 / (maxx - minx);
		scaley = 1.0 / (maxy - miny);

		printf("LOWER LIMITS:   X = %2.6f Y = %2.6f\n", minx, miny);
		printf("UPPER LIMITS:   X = %2.6f Y = %2.6f\n", maxx, maxy);
		printf("SCALING UVS BY: X = %2.6f Y = %2.6f\n\n", scalex, scaley);
		
		if (polygon_count > 0)
		{
			shading_polygons = new polygon[polygon_count];
			
			for (int i = 0; i < polygon_count; i++)
			{
				shading_polygons[i] = polygons[i];

				shading_polygons[i].uv1.x = (polygons[i].uv1.x - minx) * scalex;
				shading_polygons[i].uv1.y = (polygons[i].uv1.y - miny) * scaley;

				shading_polygons[i].uv2.x = (polygons[i].uv2.x - minx) * scalex;
				shading_polygons[i].uv2.y = (polygons[i].uv2.y - miny) * scaley;

				shading_polygons[i].uv3.x = (polygons[i].uv3.x - minx) * scalex;
				shading_polygons[i].uv3.y = (polygons[i].uv3.y - miny) * scaley;
			}
		} else cout << "ERROR: Object has no polygons!" << endl;
	} else cout << "ERROR: Object has no UVs!" << endl;
}

/******************************************************************************/
// Creates display lists, initialises textures and other housekeeping tasks
/******************************************************************************/

void object::initialiseObject(colour amb, colour dif, colour spec)
{
#ifdef _DEBUG_OBJ
	cout << "initialiseObject(void);" << endl;
#endif

	UVCounterClockwiseTest();

	scaleUVs();

	buildEdgeList();

	calcAlphaMap();
	calcPositionsMap();
	calcNormalsMap();

	initialiseBitmaps(DIFFUSE);
	initialiseBitmaps(AMBIENT);
	initialiseBitmaps(SPECULAR);
	initialiseBitmaps(FINAL);

	setFlatColour(SPECULAR, spec);
	setFlatColour(AMBIENT, amb);
	setFlatColour(DIFFUSE, dif);

// initialise object for display

	if (polygons != NULL)
	{
		list_name = glGenLists(1);
		
		glNewList(list_name, GL_COMPILE);
		
		glBegin(GL_TRIANGLES);
		for (int i = 0; i < polygon_count; i++)
		{
			polygon p = polygons[i];
			
			glNormal3f(p.normal1.x, p.normal1.y, p.normal1.z);
			glTexCoord2f(p.uv1.x, p.uv1.y);
			glVertex3f(p.vertex1.x, p.vertex1.y, p.vertex1.z);
			
			glNormal3f(p.normal2.x, p.normal2.y, p.normal2.z);
			glTexCoord2f(p.uv2.x, p.uv2.y);
			glVertex3f(p.vertex2.x, p.vertex2.y, p.vertex2.z);
			
			glNormal3f(p.normal3.x, p.normal3.y, p.normal3.z);
			glTexCoord2f(p.uv3.x, p.uv3.y);
			glVertex3f(p.vertex3.x, p.vertex3.y, p.vertex3.z);
		}
		glEnd();
		
		glEndList();
		
	} else cout << "ERROR: No polygons in object!" << endl;

// initialise object for shading
	
	if (shading_polygons != NULL)
	{
		shading_list_name = glGenLists(1);
		
		glNewList(shading_list_name, GL_COMPILE);
		
		glBegin(GL_TRIANGLES);
		for (int i = 0; i < polygon_count; i++)
		{
			polygon p = shading_polygons[i];
			
			glNormal3f(p.normal1.x, p.normal1.y, p.normal1.z);
			glTexCoord2f(p.uv1.x, p.uv1.y);
			glVertex3f(p.vertex1.x, p.vertex1.y, p.vertex1.z);
			
			glNormal3f(p.normal2.x, p.normal2.y, p.normal2.z);
			glTexCoord2f(p.uv2.x, p.uv2.y);
			glVertex3f(p.vertex2.x, p.vertex2.y, p.vertex2.z);
			
			glNormal3f(p.normal3.x, p.normal3.y, p.normal3.z);
			glTexCoord2f(p.uv3.x, p.uv3.y);
			glVertex3f(p.vertex3.x, p.vertex3.y, p.vertex3.z);
		}
		glEnd();

		glEndList();
	} else cout << "ERROR: No shading polygons in object!" << endl;

	if (CALC_AMBIENT_MAP == TRUE && CALC_DIFFUSE_MAP == TRUE && CALC_SPECULAR_MAP == TRUE)
	{
		calcAllMaps(_LIGHTS, _CAMERA, _GLOBAL_AMBIENT);
	} else {
		if (CALC_AMBIENT_MAP == TRUE) calcAmbientMap(_GLOBAL_AMBIENT);
		if (CALC_DIFFUSE_MAP == TRUE) calcDiffuseMap(_LIGHTS, _CAMERA);
		if (CALC_SPECULAR_MAP == TRUE) calcSpecularMap(_LIGHTS, _CAMERA);

		combineColourComponents();
	}

// Generates OpenGL texture objects

	glGenTextures(1, &shading_tex_name);
	glBindTexture(GL_TEXTURE_2D, shading_tex_name);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexEnvi(GL_TEXTURE_2D, GL_RGB, GL_REPLACE);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, final_colour->x, final_colour->y,
		0, GL_RGB, GL_UNSIGNED_BYTE, final_colour->bitmap_data);

	if (colour_map != NULL)
	{
		glGenTextures(1, &tex_name);
		glBindTexture(GL_TEXTURE_2D, tex_name);
		
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

		glTexEnvi(GL_TEXTURE_2D, GL_RGB, GL_REPLACE);

		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, colour_map->x, colour_map->y,
			0, GL_RGB, GL_UNSIGNED_BYTE, colour_map->bitmap_data);
	}
}

/******************************************************************************/
// Displays an object using its lighting texture, and makes calls to the calcLighing
// functions if need be
/******************************************************************************/

void object::displayShadedObject(void)
{
#ifdef _DEBUG_OBJ
	cout << "displayShadedObject(void);" << endl;
#endif

	if (visible == TRUE)
	{
		if (visible_in_camera == TRUE)
		{
			glPushMatrix();
			glMultMatrixd(transformation_matrix.m);

			visibilityTest(_CAMERA);

			glEnable(GL_TEXTURE_2D);

			glBindTexture(GL_TEXTURE_2D, shading_tex_name);

			if (CALC_AMBIENT_MAP_REALTIME == TRUE && CALC_DIFFUSE_MAP_REALTIME == TRUE && CALC_SPECULAR_MAP_REALTIME == TRUE)
			{
				calcAllMaps(_LIGHTS, _CAMERA, _GLOBAL_AMBIENT);
				
				glTexEnvi(GL_TEXTURE_2D, GL_RGB, GL_REPLACE);
				
				glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, final_colour->x, final_colour->y,
					0, GL_RGB, GL_UNSIGNED_BYTE, final_colour->bitmap_data);
			} else if (CALC_AMBIENT_MAP_REALTIME == TRUE || CALC_DIFFUSE_MAP_REALTIME == TRUE || CALC_SPECULAR_MAP_REALTIME == TRUE)
			{
				if (CALC_AMBIENT_MAP_REALTIME == TRUE) calcAmbientMap(_GLOBAL_AMBIENT);
				if (CALC_DIFFUSE_MAP_REALTIME == TRUE) calcDiffuseMap(_LIGHTS, _CAMERA);
				if (CALC_SPECULAR_MAP_REALTIME == TRUE) calcSpecularMap(_LIGHTS, _CAMERA);

				combineColourComponents();

				glTexEnvi(GL_TEXTURE_2D, GL_RGB, GL_REPLACE);

				glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, final_colour->x, final_colour->y,
					0, GL_RGB, GL_UNSIGNED_BYTE, final_colour->bitmap_data);
			}

			glCallList(shading_list_name);
			glPopMatrix();
		}
	}
}

/******************************************************************************/
// Draws the objects with colour and texture only, no shading of any kind
/******************************************************************************/

void object::displayTexturedObject(void)
{
#ifdef _DEBUG_OBJ
	cout << "displayTexturedObject(void);" << endl;
#endif

	if (visible == TRUE)
	{
		if (visible_in_camera == TRUE)
		{
			glDisable(GL_LIGHTING);
			glPushMatrix();
			glMultMatrixd(transformation_matrix.m);
			
			if (colour_map != NULL)
			{
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, tex_name);
			} else {
				glDisable(GL_TEXTURE_2D);
				colour c = diffuse.getColour(0, 0);
				glColor3f(c.r * (1.0/255.0), c.g * (1.0/255.0), c.b * (1.0/255.0));
			}
			
			glCallList(list_name);
			glPopMatrix();
		}
	}
}

/******************************************************************************/
// Renders an object using the normal OpenGL system with lighting and textures
/******************************************************************************/

void object::displayOpenGLShadedObject(void)
{
#ifdef _DEBUG_OBJ
	cout << "displayOpenGLShadedObject(void);" << endl;
#endif

	if (visible == TRUE)
	{
		if (visible_in_camera == TRUE)
		{
			glPushMatrix();
			glMultMatrixd(transformation_matrix.m);

			float dif[4], amb[4], spec[4], n(1.0 / 255.0);
			colour c = diffuse.getColour(0, 0);

			dif[0] = c.r * n; dif[1] = c.g * n; dif[2] = c.b * n; dif[3] = 1.0;
			c = ambient.getColour(0, 0);
			amb[0] = c.r * n; amb[1] = c.g * n; amb[2] = c.b * n; amb[3] = 1.0;
			c = specular.getColour(0, 0);
			spec[0] = c.r * n; spec[1] = c.g * n; spec[2] = c.b * n; spec[3] = 1.0;

			glMaterialfv(GL_FRONT, GL_DIFFUSE, dif);
			glMaterialfv(GL_FRONT, GL_AMBIENT, amb);
			glMaterialfv(GL_FRONT, GL_SPECULAR, spec);
			glMaterialf(GL_FRONT, GL_SHININESS, specular_coefficient);

			if (colour_map != NULL)
			{
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, tex_name);
				glTexEnvi(GL_TEXTURE_2D, GL_RGB, GL_BLEND);
			} else {
				glDisable(GL_TEXTURE_2D);
			}
			
			glCallList(list_name);
			glPopMatrix();
		}
	}
}

object* loadOBJfile(char*, int, int);

/******************************************************************************/
// Loads a text file that has information on the scene (object file names, properties
// and textures
/******************************************************************************/

int loadSceneFile(char *scene_file_name, object ***new_object_list)
{
#ifdef _DEBUG_OBJ
	cout << "loadSceneFile(" << scene_file_name << ");" << endl;
#endif

	FILE *scene_file;

	scene_file = fopen(scene_file_name, "r");

	if (scene_file != NULL)
	{
		char data_string[1000], object_file_name[256];
		int object_count(0), x_size, y_size;
		colour amb(50, 50, 50, 255), dif(200, 200, 200, 255), spec(255, 255, 255, 255);
		float bump, coef;
		char bump_map_name[1000], colour_map_name[1000];

		while (!feof(scene_file))
		{
			fgets(data_string, 1000, scene_file);
			object_count++;
		}

		rewind(scene_file);

		*new_object_list = new object* [object_count];

		for (int i = 0; i < object_count; i++)
		{
			fgets(data_string, 1000, scene_file);
			sscanf(data_string, "%s %d %d %d %d %d %d %d %d %d %f %s %f %d %d %s",
				object_file_name,
				&amb.r, &amb.g, &amb.b,
				&dif.r, &dif.g, &dif.b,
				&spec.r, &spec.g, &spec.b,
				&coef, bump_map_name, &bump, &x_size, &y_size, colour_map_name);

			cout << "*******************************************************" << endl;

			(*new_object_list)[i] = loadOBJfile(object_file_name, x_size, y_size);
			(*new_object_list)[i]->bumpiness = bump;
			(*new_object_list)[i]->specular_coefficient = coef;

			printf("MATERIAL PROPERTIES:\n");
			printf("AMBIENT: %3d, %3d, %3d | DIFFUSE: %3d, %3d, %3d | SPECULAR %3d, %3d, %3d\n\n", amb.r, amb.g, amb.b, dif.r, dif.g, dif.b, spec.r, spec.g, spec.b);
			printf("TEXTURES:\n");
			printf("COLOUR TEXTURE: %s | BUMP MAP TEXTURE: %s\n\n", colour_map_name, bump_map_name);
			printf("COEFFICIENTS:\n");
			printf("SPECULAR COEFFICIENT: %2.6f | BUMP COEFFICIENT: %2.6f\n\n", coef, bump);

			if (strcmp(bump_map_name, "NULL") != 0)
			{
				cout << "Setting bumpmap '" << bump_map_name << "'!" << endl;
				bitmap *new_bump;
				new_bump = LoadBMPFile(bump_map_name);
				(*new_object_list)[i]->setBumpMap(new_bump);
			}

			if (strcmp(colour_map_name, "NULL") != 0)
			{
				cout << "Setting colour map '" << colour_map_name << "'!" << endl;
				bitmap *new_colour;
				new_colour = LoadBMPFile(colour_map_name);
				(*new_object_list)[i]->setColourMap(new_colour);
			}

			(*new_object_list)[i]->initialiseObject(amb, dif, spec);
		}

		return object_count;

	} else printf("ERROR: Scene file '%s' does not exist!\n", scene_file_name);

	return 0;
}

unsigned char WriteBMPFile(char*, bitmap*);

/******************************************************************************/
// Saves the texture buffers (Polygon, Position, Normal, Ambient, Diffuse, Specular and
// Final as bitmaps
/******************************************************************************/

void object::saveTextureBuffer(int which, char *filename)
{
#ifdef _DEBUG_OBJ
	cout << "saveTextureBuffer(" << which << ", " << filename << ");" << endl;
#endif

	int x, y;
	colour c;
	bitmap *temp = NULL;

	switch(which)
	{
	case DIFFUSE:
		WriteBMPFile(filename, diffuse_colour);
		break;
	case AMBIENT:
		WriteBMPFile(filename, ambient_colour);
		break;
	case SPECULAR:
		WriteBMPFile(filename, specular_colour);
		break;
	case FINAL:
		WriteBMPFile(filename, final_colour);
		break;
	case ALPHA:
		x = final_colour->x;
		y = final_colour->y;

		temp = new bitmap(x, y, 3);
		
		for (int j = 0; j < y; j++) {
			for (int i = 0; i < x; i++) {
				int a(0);
				if (alpha_map[i][j] != NULL) a = 255;
				c.r = a;
				c.g = a;
				c.b = a;
				temp->putColour(i, j, c); } }
		
		WriteBMPFile(filename, temp);

		break;

	case NORMALS:
		cout << "Writing Normals Map" << endl;
		x = final_colour->x;
		y = final_colour->y;

		temp = new bitmap(x, y, 3);
		
		for (int j = 0; j < x; j++) {
			for (int i = 0; i < x; i++) {
				c.r = 127 * -normals_map[j][i].x + 127;
				c.g = 127 * normals_map[j][i].y + 127;
				c.b = 127 * -normals_map[j][i].z + 127;
				temp->putColour(i, j, c); } }

		WriteBMPFile(filename, temp);

		break;

	case POSITIONS:
		x = final_colour->x;
		y = final_colour->y;
		
		temp = new bitmap(x, y, 3);
		
		for (int j = 0; j < x; j++) {
			for (int i = 0; i < x; i++) {
				c.r = 127 * -position_map[j][i].x + 127;
				c.g = 127 * position_map[j][i].y + 127;
				c.b = 127 * -position_map[j][i].z + 127;
				temp->putColour(i, j, c); } }

		WriteBMPFile(filename, temp);
		
		break;
	default:
		printf("Enter valid buffer for writing to BMP!\n");
		break;
	}

	if (temp != NULL) {
		if (temp->bitmap_data != NULL) delete [] temp->bitmap_data;
		delete [] temp;
	}
}

/******************************************************************************/
// Is the edge 'tester' already in the edge_list (going either direction). Returns
// True or False
/******************************************************************************/

unsigned char isDuplicateEdge(edge tester, edge *edge_list, int count)
{
	for (int i = 0; i < count; i++)
	{
		if (edge_list[i].vertex1 == tester.vertex1)
		{
			if (edge_list[i].vertex2 == tester.vertex2)
			{
				return TRUE;
			}
		} else if (edge_list[i].vertex2 == tester.vertex1)
		{
			if (edge_list[i].vertex1 == tester.vertex2)
			{
				return TRUE;
			}
		}
	}

	return FALSE;
}

/******************************************************************************/
// Determines which polygons share this edge and returns its index (-1 if none do)
/******************************************************************************/

int sharesEdge(int tester, edge test_edge, polygon *polygon_list, int count)
{
	if (tester < count)
	{
		for (int i = 0; i < count; i++)
		{
			if (i != tester)
			{
				if (polygon_list[i].vertex1 == test_edge.vertex1)
				{
					if (polygon_list[i].vertex2 == test_edge.vertex2)
					{
						return i;
					}
				}
				
				if (polygon_list[i].vertex2 == test_edge.vertex1)
				{
					if (polygon_list[i].vertex1 == test_edge.vertex2)
					{
						return i;
					}
				}

				if (polygon_list[i].vertex2 == test_edge.vertex1)
				{
					if (polygon_list[i].vertex3 == test_edge.vertex2)
					{
						return i;
					}
				}
				
				if (polygon_list[i].vertex3 == test_edge.vertex1)
				{
					if (polygon_list[i].vertex2 == test_edge.vertex2)
					{
						return i;
					}
				}

				if (polygon_list[i].vertex3 == test_edge.vertex1)
				{
					if (polygon_list[i].vertex1 == test_edge.vertex2)
					{
						return i;
					}
				}
				
				if (polygon_list[i].vertex1 == test_edge.vertex1)
				{
					if (polygon_list[i].vertex3 == test_edge.vertex2)
					{
						return i;
					}
				}
			}
		}

		return -1;
	}

	return -1;
}

/******************************************************************************/
// Builds an edge list with polygon sharing information and edge visibility information
/******************************************************************************/

void object::buildEdgeList(void)
{
	int count = 0, visible_edges_count = 0, boundary_edges = 0;

	edge *temp_edges = new edge [polygon_count * 3];
	edge temp;

	for (int i = 0; i < polygon_count; i++)
	{
		polygons[i].plane_normal.normalise();
	}

	for (int i = 0; i < polygon_count; i++)
	{
// Creates a temporary edge and determines whether it already exists or not

		temp.vertex1 = polygons[i].vertex1;
		temp.vertex2 = polygons[i].vertex2;
		temp.visible = FALSE;

		if (isDuplicateEdge(temp, temp_edges ,count) == FALSE)
		{
// Sets edge sharing information
			temp.left_polygon = i;
			temp.right_polygon = sharesEdge(i, temp, polygons, polygon_count);

// If the edge is shared by two polygons, then the angle between them is tested. If it
// is under a certain threshold, then the edge is deem sharp and therefore visible in
// the toon renderer
// If the edge is only used by one polygon, then it is visible as it is a boundary/outline
// edge

			if (temp.right_polygon != -1)
			{
				double d;

				d = dot(polygons[temp.left_polygon].plane_normal, polygons[temp.right_polygon].plane_normal);

				if (d <= sharp_edge_angle)
				{
					visible_edges_count++;
					temp.visible = TRUE;
				}
			} else {
				boundary_edges++;
				visible_edges_count++;
				temp.visible = TRUE;
			}

			temp_edges[count] = temp;
			count++;
		}

// The steps above are repeated for all three edges of every polygon

		temp.vertex1 = polygons[i].vertex2;
		temp.vertex2 = polygons[i].vertex3;
		temp.visible = FALSE;

		if (isDuplicateEdge(temp, temp_edges ,count) == FALSE)
		{
			temp.left_polygon = i;
			temp.right_polygon = sharesEdge(i, temp, polygons, polygon_count);

			if (temp.right_polygon != -1)
			{
				double d;

				d = dot(polygons[temp.left_polygon].plane_normal, polygons[temp.right_polygon].plane_normal);

				if (d <= sharp_edge_angle)
				{
					visible_edges_count++;
					temp.visible = TRUE;
				}
			} else {
				boundary_edges++;
				visible_edges_count++;
				temp.visible = TRUE;
			}

			temp_edges[count] = temp;
			count++;
		}

		temp.vertex1 = polygons[i].vertex3;
		temp.vertex2 = polygons[i].vertex1;
		temp.visible = FALSE;

		if (isDuplicateEdge(temp, temp_edges ,count) == FALSE)
		{
			temp.left_polygon = i;
			temp.right_polygon = sharesEdge(i, temp, polygons, polygon_count);

			if (temp.right_polygon != -1)
			{
				double d;

				d = dot(polygons[temp.left_polygon].plane_normal, polygons[temp.right_polygon].plane_normal);

				if (d <= sharp_edge_angle)
				{
					visible_edges_count++;
					temp.visible = TRUE;
				}
			} else {
				boundary_edges++;
				visible_edges_count++;
				temp.visible = TRUE;
			}

			temp_edges[count] = temp;
			count++;
		}
		float percent, a;
		a = 100.0 / polygon_count;
		percent = i * a;
		printf("Generating Edge List: %d\%\r", (int)percent);
	}

	if (count > 0)
	{
		edge_count = count;

		edges = new edge [count];

		for (int i = 0; i < count; i++) edges[i] = temp_edges[i];
	}

	delete [] temp_edges;

	printf("\nEDGE LIST GENERATED:\n");
	printf("TOTAL NUMBER OF EDGES:    %4d\n", count);
	printf("NUMBER OF SHARP EDGES:    %4d\n", visible_edges_count);
	printf("NUMBER OF BOUNDARY EDGES: %4d\n\n", boundary_edges);
}

#define STEPS 3

/******************************************************************************/
// Renders an object in toon shaded mode.
// the polygons are rendered first with flattened normals
// the edges are rendered afterwards over the polygons
/******************************************************************************/

void object::displayToonShadedObject(void)
{
	double n = 1.0 / 255.0, lr, lg, lb, difr, difg, difb, d;
	colour c = diffuse.getColour(0, 0), a = ambient.getColour(0, 0), s = specular.getColour(0, 0);
	light *cur_light;
	vector dir;
	float dif[4], amb[4], spec[4];

	visibilityTest(_CAMERA);

	glShadeModel(GL_SMOOTH);
	glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(1.0, 2.0);
	glDisable(GL_CULL_FACE);

	glEnable(GL_LIGHTING);
	
	dif[0] = c.r * n;
	dif[1] = c.g * n;
	dif[2] = c.b * n;
	dif[3] = 1.0;

	amb[0] = a.r * n;
	amb[1] = a.g * n;
	amb[2] = a.b * n;
	amb[3] = 1.0;
	
	spec[0] = s.r * n;
	spec[1] = s.g * n;
	spec[2] = s.b * n;
	spec[3] = 1.0;

	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, dif);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, amb);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, spec);

	glColor3fv(dif);

// Renders the objects polygons

	glBegin(GL_TRIANGLES);

	for (int i = 0; i < polygon_count; i++)
	{
		if (polygons[i].visible == 1)
		{
			vector cur_normal1, cur_normal2, cur_normal3, light_dir;
			double d, t;

			cur_light = _LIGHTS;

			cur_normal1 = 0.0;
			cur_normal2 = 0.0;
			cur_normal3 = 0.0;

// Clamp polygons normals to point towards the light if the angle between the original
// normal and the light vector falls below a certain threshold

			while (cur_light != NULL)
			{
				light_dir = cur_light->getLightDirection(polygons[i].vertex1);
				light_dir.normalise();

				d = dot(polygons[i].normal1, light_dir);

				if (d > 0.5) cur_normal1 = cur_normal1 + light_dir;
				else {
					t = d * 2.0;
					cur_normal1 = polygons[i].normal1 * (1.0 - t) + light_dir * t;
				}

				light_dir = cur_light->getLightDirection(polygons[i].vertex2);				light_dir.normalise();
				light_dir.normalise();

				d = dot(polygons[i].normal2, light_dir);

				if (d > 0.5) cur_normal2 = cur_normal2 + light_dir;
				else {
					t = d * 2.0;
					cur_normal2 = polygons[i].normal2 * (1.0 - t) + light_dir * t;
				}

				light_dir = cur_light->getLightDirection(polygons[i].vertex3);
				light_dir.normalise();

				d = dot(polygons[i].normal3, light_dir);

				if (d > 0.5) cur_normal3 = cur_normal3 + light_dir;
				else {
					t = d * 2.0;
					cur_normal3 = polygons[i].normal3 * (1.0 - t) + light_dir * t;
				}

				cur_light = cur_light->nextptr;
			}

			cur_normal1.normalise();
			cur_normal2.normalise();
			cur_normal3.normalise();

			glNormal3f(cur_normal1.x, cur_normal1.y, cur_normal1.z);
			glVertex3f(polygons[i].vertex1.x, polygons[i].vertex1.y, polygons[i].vertex1.z);

			glNormal3f(cur_normal2.x, cur_normal2.y, cur_normal2.z);
			glVertex3f(polygons[i].vertex2.x, polygons[i].vertex2.y, polygons[i].vertex2.z);

			glNormal3f(cur_normal3.x, cur_normal3.y, cur_normal3.z);
			glVertex3f(polygons[i].vertex3.x, polygons[i].vertex3.y, polygons[i].vertex3.z);
		}
	}
	glEnd();

	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);
	glColor3f(dif[0] / 3.0, dif[1] / 3.0, dif[2] / 3.0);

//#define DRAW_NORMALS

#ifdef DRAW_NORMALS

	glBegin(GL_LINES);

	for (int i = 0; i < polygon_count; i++)
	{
		if (polygons[i].visible == 1)
		{
			vector p = (polygons[i].vertex1 + polygons[i].vertex2 + polygons[i].vertex3) / 3;
			vector n = p + polygons[i].plane_normal;

			glVertex3f(p.x, p.y, p.z);
			glVertex3f(n.x, n.y, n.z);
		}
	}
	glEnd();
#endif

// Draws the objects visible edges (determined during the building of the edge list)
// and its outline edges (if the faces that share the edge are flagged differently following
// a visibility test, then the edge is part of the outline)

	glLineWidth(2.0);

	glDepthFunc(GL_LEQUAL);

	glBegin(GL_LINES);

	for (int i = 0; i < edge_count; i++)
	{
		if (edges[i].visible == TRUE)
		{
			glVertex3f(edges[i].vertex1.x, edges[i].vertex1.y, edges[i].vertex1.z);
			glVertex3f(edges[i].vertex2.x, edges[i].vertex2.y, edges[i].vertex2.z);
		} else if (edges[i].right_polygon != -1)
		{
			if (polygons[edges[i].left_polygon].visible != polygons[edges[i].right_polygon].visible)
			{
				glVertex3f(edges[i].vertex1.x, edges[i].vertex1.y, edges[i].vertex1.z);
				glVertex3f(edges[i].vertex2.x, edges[i].vertex2.y, edges[i].vertex2.z);
			}
		}
	}
	
	glEnd();

	glEnable(GL_CULL_FACE);
}
