// object.h: interface for the object class.
//
//////////////////////////////////////////////////////////////////////

#ifndef _OBJECT_H_
#define _OBJECT_H_

#include "texture.h"
#include "bitmap.h"
#include "polygon.h"
#include "edge.h"
#include "flat_colour.h"
#include "colour_texture.h"
#include "light.h"

class object  
{
public:
	object(int = 0, int = 0);
	virtual ~object();

	void calcAlphaMap(void);
	void calcPositionsMap(void);
	void calcNormalsMap(void);
	void initialiseBitmaps(int);

	void setFlatColour(int, colour);
	void setColourTexture(int, bitmap);

	void setBumpMap(bitmap*);
	void setColourMap(bitmap*);

	void calcAmbientMap(colour);
	void calcDiffuseMap(light*, vector);
	void calcSpecularMap(light*, vector);
	void calcAllMaps(light*, vector, colour);

	void combineColourComponents(void);

	void visibilityTest(vector);
	void UVCounterClockwiseTest(void);
	void scaleUVs(void);
	void buildEdgeList(void);

	void initialiseObject(colour, colour, colour);
	void displayShadedObject(void);
	void displayTexturedObject(void);
	void displayToonShadedObject(void);
	void displayOpenGLShadedObject(void);

	void saveTextureBuffer(int, char*);

	vector *vertices;
	vector *normals;
	vector *uvs;
	polygon *polygons;
	polygon *shading_polygons;
	edge *edges;

	int vertex_count;
	int uv_count;
	int normals_count;
	int polygon_count;
	int edge_count;

	matrix transformation_matrix;

	unsigned int list_name;
	unsigned int tex_name;
	unsigned int shading_list_name;
	unsigned int shading_tex_name;

	texture diffuse;
	texture ambient;
	texture specular;
	texture normal;

	polygon ***alpha_map;

	bitmap *bump_map;
	bitmap *colour_map;

	bitmap *diffuse_colour;
	bitmap *ambient_colour;
	bitmap *diffuse_and_ambient_colour;
	bitmap *specular_colour;
	bitmap *final_colour;

	double specular_coefficient;
	double sharp_edge_angle;
	double bumpiness;

	unsigned char visible;
	unsigned char visible_in_camera;

	vector **normals_map;
	vector **position_map;
};

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

object::object(int x, int y)
{
	vertices = NULL;
	normals = NULL;
	uvs = NULL;
	polygons = NULL;
	shading_polygons = NULL;
	edges = NULL;

	edge_count = 0;

	list_name = 0;
	tex_name = 0;
	shading_list_name = 0;
	shading_tex_name = 0;

	specular_coefficient = 100.0;
	sharp_edge_angle = 0.707;
	bumpiness = 1.0;

	visible = 1;
	visible_in_camera = 1;

	if (x > 0 && y > 0)
	{
		alpha_map = new polygon** [x];
		for (int i = 0; i < x; i++)
		{
			alpha_map[i] = new polygon* [y];
			for (int j = 0; j < y; j++)
			{
				alpha_map[i][j] = NULL;
			}
		}
	}

	bump_map = NULL;
	colour_map = NULL;

	diffuse_colour = new bitmap(x, y, 3);
	ambient_colour = new bitmap(x, y, 3);
	diffuse_and_ambient_colour = new bitmap(x, y, 3);
	specular_colour = new bitmap(x, y, 3);
	final_colour = new bitmap(x, y, 3);

	if (x != 0 && y != 0)
	{
		int i;

		normals_map = new vector* [x];
		for (i = 0; i < x; i++) normals_map[i] = new vector[y];

		position_map = new vector* [x];
		for (i = 0; i < x; i++) position_map[i] = new vector[y];
	} else {
		normals_map = NULL;
		position_map = NULL; }
}

object::~object()
{

}

#endif
