// polygon.cpp: implementation of the polygon class.
//
//////////////////////////////////////////////////////////////////////

#include "polygon.h"

/******************************************************************************/
// Returns true if the point is inside the polygon. Uses the 'left of line' rule
/******************************************************************************/

unsigned char polygon::isPointInsidePolygon(vector point)
{
	vector v, n, p;
	double d;

	v = vertex2 - vertex1;
	v.normalise();

	p = point - vertex1;

	n.x = -v.y;
	n.y = v.x;

	d = dot(n, p);

	if (d >= -1.0) {
		v = vertex3 - vertex2;
		v.normalise();

		p = point - vertex2;

		n.x = -v.y;
		n.y = v.x;

		d = dot(n, p);

		if (d >= -1.0) {
			v = vertex1 - vertex3;
			v.normalise();

			p = point - vertex3;

			n.x = -v.y;
			n.y = v.x;

			d = dot(n, p);

			if (d >= -1.0) return 1; } }

	return 0;
}

/******************************************************************************/
// interpolates between UV's and Vertices to find the point in space that the
// UV's correspond to
/******************************************************************************/

vector polygon::uvSpaceToWorldSpace(vector pos)
{
	double v1, v2, v3;
	vector p1, p2, p3, x1, x2, x3, vt1, vt2, vt3, pt1, pt2, pt3;

	vt1 = uv1;
	vt2 = uv2;
	vt3 = uv3;
	pt1 = vertex1;
	pt2 = vertex2;
	pt3 = vertex3;

	double t1 = vt1.y - vt2.y, t2 = vt3.y - vt2.y;

	while (t1 == 0.0 || t2 == 0.0)
	{
		vector tempv = vt1, tempp = pt1;
		vt1 = vt2;
		vt2 = vt3;
		vt3 = tempv;

		pt1 = pt2;
		pt2 = pt3;
		pt3 = tempp;

		t1 = vt1.y - vt2.y;
		t2 = vt3.y - vt2.y;
	}

	v1 = pos.y - vt2.y;
	v2 = vt1.y - vt2.y;
	v3 = vt1.y - pos.y;

	if (v2 != 0.0)
	{
		p1 = vt1 * (v1 / v2) + vt2 * (v3 / v2);
		x1 = pt1 * (v1 / v2) + pt2 * (v3 / v2);
		
		v1 = pos.y - vt2.y;
		v2 = vt3.y - vt2.y;
		v3 = vt3.y - pos.y;
		
		if (v2 != 0.0)
		{
			p2 = vt3 * (v1 / v2) + vt2 * (v3 / v2);
			x2 = pt3 * (v1 / v2) + pt2 * (v3 / v2);

			v1 = p2.x - pos.x;
			v2 = p2.x - p1.x;
			v3 = pos.x - p1.x;

			if (v2 != 0.0)
			{
				x3 = x1 * (v1 / v2) + x2 * (v3 / v2);
			}// else cout << "fail on 3rd" << endl;
		}// else cout << "fail on 2nd" << endl;
	}// else cout << "fail on 1st" << endl;

	return x3;
}

/******************************************************************************/
// Returns the normal on the surface in world space that corresponds to a
// point in UV space
/******************************************************************************/

vector polygon::getNormalAtUV(vector pos)
{
	double v1, v2, v3;
	vector p1, p2, p3, n1, n2, n3, vt1, vt2, vt3, nt1, nt2, nt3;

	vt1 = uv1;
	vt2 = uv2;
	vt3 = uv3;
	nt1 = normal1;
	nt2 = normal2;
	nt3 = normal3;

	double t1 = vt1.y - vt2.y, t2 = vt3.y - vt2.y;

	while (t1 == 0.0 || t2 == 0.0)
	{
		vector tempv = vt1, tempn = nt1;
		vt1 = vt2;
		nt1 = nt2;

		vt2 = vt3;
		nt2 = nt3;

		vt3 = tempv;
		nt3 = tempn;

		t1 = vt1.y - vt2.y;
		t2 = vt3.y - vt2.y;
	}

	v1 = pos.y - vt2.y;
	v2 = vt1.y - vt2.y;
	v3 = vt1.y - pos.y;

	if (v2 != 0.0)
	{
		p1 = vt1 * (v1 / v2) + vt2 * (v3 / v2);
		n1 = nt1 * (v1 / v2) + nt2 * (v3 / v2);
		
		v1 = pos.y - vt2.y;
		v2 = vt3.y - vt2.y;
		v3 = vt3.y - pos.y;
		
		if (v2 != 0.0)
		{
			p2 = vt3 * (v1 / v2) + vt2 * (v3 / v2);
			n2 = nt3 * (v1 / v2) + nt2 * (v3 / v2);
			
			v1 = p2.x - pos.x;
			v2 = p2.x - p1.x;
			v3 = pos.x - p1.x;
			
			if (v2 != 0.0)
			{
				n3 = n1 * (v1 / v2) + n2 * (v3 / v2);
			}// else cout << "fail on 3rd" << endl;
		}// else cout << "fail on 2nd" << endl;
	}// else cout << "fail on 1st" << endl;

	return n3;
}