// vector.cpp: implementation of the vector class.
//
//////////////////////////////////////////////////////////////////////

#include "vector.h"

vector vector::operator-()
{
	vector n;

	n.x = 0 - x;
	n.y = 0 - y;
	n.z = 0 - z;

	return n;
}

vector vector::operator+(vector right)
{
	vector n;

	n.x = x + right.x;
	n.y = y + right.y;
	n.z = z + right.z;

	return n;
}

vector vector::operator-(vector right)
{
	vector n;

	n.x = x - right.x;
	n.y = y - right.y;
	n.z = z - right.z;

	return n;
}

vector vector::operator*(vector right)
{
	vector n;

	n.x = x * right.x;
	n.y = y * right.y;
	n.z = z * right.z;

	return n;
}

vector vector::operator/(vector right)
{
	vector n;

	n.x = x / right.x;
	n.y = y / right.y;
	n.z = z / right.z;

	return n;
}

vector vector::operator+(double right)
{
	vector n;

	n.x = x + right;
	n.y = y + right;
	n.z = z + right;

	return n;
}

vector vector::operator-(double right)
{
	vector n;

	n.x = x - right;
	n.y = y - right;
	n.z = z - right;

	return n;
}

vector vector::operator*(double right)
{
	vector n;

	n.x = x * right;
	n.y = y * right;
	n.z = z * right;

	return n;
}

vector vector::operator/(double right)
{
	vector n;

	n.x = x / right;
	n.y = y / right;
	n.z = z / right;

	return n;
}

vector vector::operator*(matrix right)
{
	vector n;

	n.x = x * right.m[0] + y * right.m[4] + z * right.m[8]  + w * right.m[12];
	n.y = x * right.m[1] + y * right.m[5] + z * right.m[9]  + w * right.m[13];
	n.z = x * right.m[2] + y * right.m[6] + z * right.m[10] + w * right.m[14];
	n.w = x * right.m[3] + y * right.m[7] + z * right.m[11] + w * right.m[15];

	return n;
}

vector vector::operator%(vector right)
{
	vector n;

	n.x = (y * right.z) - (z * right.y);
	n.y = (z * right.x) - (x * right.z);
	n.z = (x * right.y) - (y * right.x);

	return n;
}

ostream& operator<<(ostream &output, vector n)
{
	output << "[" << n.x << " " << n.y << " " << n.z << "]";
	return output;
}

unsigned char vector::operator==(vector right)
{
	return ((x == right.x) && (y == right.y) && (z == right.z) && (w == right.w));
}

unsigned char vector::operator!=(vector right)
{
	return !((x == right.x) && (y == right.y) && (z == right.z) && (w == right.w));
}

void vector::operator()(double a, double b, double c)
{
	x = a;
	y = b;
	z = c;
}

void vector::operator+=(vector right)
{
	*this = *this + right;
}

void vector::operator-=(vector right)
{
	*this = *this - right;
}

void vector::operator*=(vector right)
{
	*this = *this * right;
}

void vector::operator/=(vector right)
{
	*this = *this + right;
}

void vector::operator+=(double right)
{
	*this = *this + right;
}

void vector::operator-=(double right)
{
	*this = *this + right;
}

void vector::operator*=(double right)
{
	*this = *this + right;
}

void vector::operator/=(double right)
{
	*this = *this + right;
}


inline double vector::mag(void)
{
	return sqrt((x*x) + (y*y) + (z*z));
}

vector vector::normalised(void)
{
	double length;
	vector new_vector = *this;

	length = mag();

	if (length != 0) {
		if (new_vector.x != 0) new_vector.x /= length;
		if (new_vector.y != 0) new_vector.y /= length;
		if (new_vector.z != 0) new_vector.z /= length; }

	return new_vector;
}

inline void vector::normalise(void)
{
	double length;

	length = mag();

	if (length != 0) {
		if (x != 0) x /= length;
		if (y != 0) y /= length;
		if (z != 0) z /= length; }
}

inline void vector::setlength(double new_length)
{
	vector new_vector;

	new_vector = normalised();

	x = new_vector.x * new_length;
	y = new_vector.y * new_length;
	z = new_vector.z * new_length;
}

inline double dot(vector v1, vector v2)
{
	return (v1.x * v2.x) + (v1.y * v2.y) + (v1.z * v2.z);
}

unsigned char isPointInsidePolygon(vector point, vector vertex1, vector vertex2, vector vertex3)
{
	vector v, n, p;
	double d;

	v = vertex2 - vertex1;
	v.normalise();

	p = point - vertex1;

	n.x = -v.y;
	n.y = v.x;

	d = dot(n, p);

	if (d >= 0.0) {
		v = vertex3 - vertex2;
		v.normalise();

		p = point - vertex2;

		n.x = -v.y;
		n.y = v.x;

		d = dot(n, p);

		if (d >= 0.0) {
			v = vertex1 - vertex3;
			v.normalise();

			p = point - vertex3;

			n.x = -v.y;
			n.y = v.x;

			d = dot(n, p);

			if (d >= 0.0) return 1; } }

	return 0;
}